; ACL2 Version 1.9 -- A Computational Logic for Applicative Common Lisp
; Copyright (C) 1997  Computational Logic, Inc.

; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.

; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

; Written by:  Matt Kaufmann               and J Strother Moore
; email:       Matt_Kaufmann@email.mot.com and Moore@cs.utexas.edu
; Computational Logic, Inc.
; 1717 West Sixth Street, Suite 290
; Austin, TX 78703-4776 U.S.A.

; Because of the last form in this file, this file should not be loaded as part
; of an executable ACL2 image.

(in-package "ACL2")

; We put the following check here to be sure that the function
; legal-variable-or-constant-namep checks for all possible lambda list
; keywords.

(dolist (name lambda-list-keywords)
        (cond

; The check from legal-variable-or-constant-namep is here.

         ((let ((s (symbol-name name)))
            (or (= (length s) 0)
                (not (eql (char s 0) #\&))))
          (error "We assume that all lambda-list-keywords start with the ~
                  character #\&.~%However, ~s does not.  ACL2 will not work in ~
                  this Common Lisp."
                 name))))

(or (and (integerp char-code-limit)
         (<= 256 char-code-limit)
         (typep 256 'fixnum))
    (error "We assume that 256 is a fixnum not exceeding char-code-limit, for ~
            purposes of~%character manipulation.  ACL2 will not work in this ~
            Common Lisp."))

;                           ACL2 CHARACTERS

; At one time we read the legal ACL2 characters from a string in a file, so
; that we could be confident that we know exactly what they are.  But if
; code-char is not a function, then we lose anyhow, because we have defined it
; in ACL2.  And if code-char is a function, then we can do our checking without
; having to get the characters from a file, the idea being that the range of
; code-char on {0,1,...,255} is exactly the set of legal ACL2 characters.

; Note that this issue is orthogonal to the problem of ensuring that #\c is
; read in as the same character in different ACL2 sessions, for any ACL2
; character c.  (Even #\Newline is potentially problematic!)  Because of the
; latter issue, our story about ACL2 should include a caveat that certified
; books may only be soundly loaded into the same Lisp image in which they were
; certified.  When we say ``same lisp image'' we don't mean the same exact
; process necessarily, but rather simply an invocation of the same saved image.
; Another reason for such a caveat, independent of the character-reading issue,
; is that different saved images may be tied into different compilers, thus
; making the object files of the books incompatible.

; Of course, we could just disallow non-ascii characters.  But, we
; would really like to have a notation for reading in arbitrary
; characters, and the easiest way to do that which is consistent with
; the standard is to allow #\c for every 8-bit character c.

; Note that the Macintosh handles characters somewhat differently from other
; implementations.  At some point we will probably modify the current treatment
; of this difficulty.  For now, we limit our claims about ACL2 by saying that
; portability of theorems is not claimed when either of the two ACL2 processes
; is running under MCL, unless they are both running the same implementation of
; MCL.

; Test that all purportedly standard characters are standard, and vice versa.

#-mcl
(dotimes (i 256)
         (let ((ch (code-char i)))
           (or (equal (standard-char-p ch)
                      (or (= i 10)
                          (and (>= i 32)
                               (<= i 126))))
               (error "This Common Lisp is unsuitable for ACL2 because ~
                      the character~%with char-code ~d ~a standard in ~
                      this Common Lisp but should~%~abe."
                      i
                      (if (standard-char-p ch) "is" "is not")
                      (if (standard-char-p ch) "not " "")))))

; Check that char-upcase and char-downcase have the same values in all
; lisps, and in particular, keep us in the realm of ACL2 characters.

#-mcl
(dotimes (i 256)
         (let ((ch (code-char i)))
           (or (eql (char-downcase ch)
                    (if (and (>= i 65)
                             (<= i 90))
                        (code-char (+ i 32))
                      ch))
               (error "This Common Lisp is unsuitable for ACL2 because ~
                      (char-downcase ~s)~%is ~s but should be ~s."
                      ch
                      (char-downcase ch)
                      (if (and (>= i 65)
                               (<= i 90))
                          (code-char (+ i 32))
                        ch)))))

#-mcl
(dotimes (i 256)
         (let ((ch (code-char i)))
           (or (eql (char-upcase ch)
                    (if (and (>= i 97)
                             (<= i 122))
                        (code-char (- (char-code ch) 32))
                      ch))
               (error "This Common Lisp is unsuitable for ACL2 because ~
                      (char-upcase ~s)~%is ~s but should be ~s."
                      ch
                      (char-upcase ch)
                      (if (and (>= i 65)
                               (<= i 90))
                          (code-char (- (char-code ch) 32))
                        ch)))))

;                           FEATURES, MISC.

; The conditional reader macro doesn't care about the package when it looks for
; the symbol.  In fact, :UNIX is not a member of *features* in gcl; LISP:UNIX
; is.

#-(or unix apple)
(error "This Common Lisp is unsuitable for ACL2 because~%~
        neither UNIX nor APPLE is a member of *features*.")

(or (typep array-dimension-limit 'fixnum)
    (error "We assume that ARRAY-DIMENSION-LIMIT is a fixnum.  CLTL2 ~
            requires this.  ACL2 will not work in this Common Lisp."))

; The following check must be put last in this file, since we don't entirely
; trust that it won't currupt the current image.

(dolist (var *copy-of-common-lisp-symbols-from-main-lisp-package*)
        (or (member var *copy-of-common-lisp-specials-and-constants* :test #'eq)
            (if (and (let ((s (symbol-name var)))
                       (or (= (length s) 0)
                           (not (eql (char s 0) #\&))))
                     (or (boundp var)
                         (eval `(let ((,var t))
                                  (declare (ignore ,var))
                                  (boundp ',var)))))
                (error "The following constant or special variable in the main~%~
                        Lisp package needs to be included in the list~%~
                        *common-lisp-specials-and-constants*:  ~s."
                       var))))

; Do not put any forms below!  See comment above.
